/*! ********************************************
 * GrayBox.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox 
{
	
public class GrayBox 
{
	protected var _eventList   :MidiEvent = null;
	protected var _processData :ProcessData = null;
	protected var _moduleSlots :int = 0;
	protected var _modules     :Vector.<IGrayBoxModule> = null;
	protected var _bypassFlag  :Vector.<Boolean> = null;
	protected var _defAudioLatency :Number = 0.0;
	protected var _audioLatency    :Number = 0.0;
// Main System ---
	public function getSampleRate():Number{ return _processData.sampleRate; }
	public function setSampleRate(sampleRate:Number):Number
	{
		if (sampleRate < 1.0)return _processData.sampleRate;
		_processData.sampleRate = sampleRate;
		for (var i:int = 0; i < _moduleSlots; i++) if (_modules[i] != null) _modules[i].updateSampleRate(sampleRate);
		return _processData.sampleRate;
	}
	public function getAudioChannels():int{ return _processData.audioChannels; }
	public function setAudioChannels(ch:int):int
	{
		if (ch < 2)return _processData.audioChannels;
		_processData.audioChannels = ch;
		for (var i:int = 0; i < _moduleSlots; i++) if (_modules[i] != null) _modules[i].updateAudioChannels(ch);
		return _processData.audioChannels;
	}
	public function getProcessBlockSize():int{ return _processData.blockSize; }
	public function setProcessBlockSize(size:int):int
	{
		if (size < 1)return _processData.blockSize;
		_processData.blockSize = size;
		for (var i:int = 0; i < _moduleSlots; i++) if (_modules[i] != null) _modules[i].updateProcessBlockSize(size);
		return _processData.blockSize;
	}
	public function getDefaultAudioLatency():Number{ return _defAudioLatency; }
	public function setDefaultAudioLatency(val:Number):void { _defAudioLatency = val; }
	public function audioLatency(val:Number):void { _audioLatency = val; }
	public function initialize():void
	{
		_defAudioLatency = 0.0;
		_audioLatency = 0.0;
		_clearProcessData();
		for (var i:int = 0; i < _moduleSlots; i++)
		{
			detachModule(i);
			_bypassFlag[i] = false;
		}
	}
// Module ---
	public function getModuleSlots():int { return _moduleSlots; }
	public function setModuleSlots(num:int):int
	{
		if (num < 1) return _moduleSlots;
		var i:int;
		for (i = _moduleSlots; i < num; i++) detachModule(i);
		_modules.fixed = false;
		_modules.length = num;
		_modules.fixed = true;
		_bypassFlag.fixed = false;
		_bypassFlag.length = num;
		_bypassFlag.fixed = true;
		for (i = _moduleSlots; i < num; i++)
		{
			_modules[i] = null;
			_bypassFlag[i] = false;
		}
		return _moduleSlots = num;
	}
	public function getModule(index:int):IGrayBoxModule
	{
		if (index >= _moduleSlots) return null;
		return _modules[index];
	}
	public function attachModule(index:int, module:IGrayBoxModule):IGrayBoxModule
	{
		if (index >= _moduleSlots) return null;
		if (module != null && module == _modules[index]) return module;
		var _mdl:IGrayBoxModule;
		for (var i:int = 0; i < _moduleSlots; i++)
		{
			if (module != _modules[i]) continue;
			_mdl = _modules[i];
			_modules[i] = null;
			_mdl.disconnectModule();
		}
		_modules[index] = module;
		module.connectModule(this);
		return module;
	}
	public function detachModule(index:int):IGrayBoxModule
	{
		if (index >= _moduleSlots) return null;
		var _mdl:IGrayBoxModule = _modules[index];
		_modules[index] = null;
		if (_mdl != null)_mdl.disconnectModule();
		return _mdl;
	}
	public function getBypassStatus(index:int):Boolean
	{
		if (index >= _moduleSlots) return false;
		return _bypassFlag[index];
	}
	public function bypassProcess(index:int, bypass:Boolean):void
	{
		if (index >= _moduleSlots) return;
		_bypassFlag[index] = bypass;
		if (_modules[index] != null)_modules[index].updateBypassStatus(bypass);
	}
	public function resetModules():void
	{
		for (var i:int = 0; i < _moduleSlots; i++) if (_modules[i] != null)_modules[i].resetModule();
	}
	public function initiModules():void
	{
		for (var i:int = 0; i < _moduleSlots; i++) if (_modules[i] != null)_modules[i].initiModule();
	}
// Process ---
	public function suspendProcess():void
	{
		for (var i:int = 0; i < _moduleSlots; i++) if (_modules[i] != null) _modules[i].suspendProcess();
		//if (_processData.processingStatus == 3)
		//{
			//_processData.processingStatus = 2;
			//for (var i:int = 0; i < _moduleSlots; i++) if (_modules[i] != null) _modules[i].updateProcessingStatus(_processData.processingStatus);
		//}
	}
	public function resumeProcess():void
	{
		for (var i:int = 0; i < _moduleSlots; i++) if (_modules[i] != null) _modules[i].resumeProcess();
		//if (_processData.processingStatus == 2)
		//{
			//_processData.processingStatus = 3;
			//for (var i:int = 0; i < _moduleSlots; i++) if (_modules[i] != null) _modules[i].updateProcessingStatus(_processData.processingStatus);
		//}
	}
	public function startProcess():void
	{
		for (var i:int = 0; i < _moduleSlots; i++) if (_modules[i] != null) _modules[i].startProcess();
		//_processData.processingStatus = 3;
		//for (var i:int = 0; i < _moduleSlots; i++) if (_modules[i] != null) _modules[i].updateProcessingStatus(_processData.processingStatus);
	}
	public function stopProcess():void
	{
		for (var i:int = 0; i < _moduleSlots; i++) if (_modules[i] != null) _modules[i].stopProcess();
		//_processData.processingStatus = 0;
		//for (var i:int = 0; i < _moduleSlots; i++) if (_modules[i] != null) _modules[i].updateProcessingStatus(_processData.processingStatus);
	}
	protected function _clearProcessData():void
	{
		_processData.midiEvents = null;
		_processData.audioOut = null;
		_processData.offsetFrames = 0;
		_processData.sampleFrames = 0;
		_audioLatency = _defAudioLatency;
	}
	public function process(audioOut:Vector.<Vector.<Number>>, offsetFrames:int, sampleFrames:int):void
	{
	// begin Process -->
		_eventList.clear();
		_processData.midiEvents = _eventList;
		_processData.audioOut = audioOut;
		_processData.offsetFrames = offsetFrames;
		_processData.sampleFrames = sampleFrames;
		_processData.audioLatency = _audioLatency;
	// Modules Process -->
		var i:int;
		//for (i = 0; i < _moduleSlots; i++) if (_modules[i] != null && !_bypassFlag[i]) _modules[i].preProcess(_processData);
		for (i = 0; i < _moduleSlots; i++) if (_modules[i] != null && !_bypassFlag[i]) _modules[i].process(_processData);
		//for (i = 0; i < _moduleSlots; i++) if (_modules[i] != null && !_bypassFlag[i]) _modules[i].postProcess																																																	+															+	+	+	+	+	+	+	+	+				3															3	3	3	3	3	3	Q11111111111111111111111111111111111111111111`````````````````````````````````````````````7YUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUTFGGGGGGGGGGGGGGGG(_processData);
	// after Process -->
		_clearProcessData();
	}
// Constructor
	public function GrayBox(blockSize:int, sampleRate:Number, audioChannels:int = 2, moduleSlots:int = 4):void
	{
		// processData
		_eventList = new MidiEvent();
		_processData = new ProcessData();
		_processData.blockSize = (blockSize < 1)? 1 : blockSize;
		_processData.sampleRate = (sampleRate < 1.0)? 1.0 : sampleRate;
		_processData.audioChannels = (audioChannels >= 2)? audioChannels : 2;
		// module
		_moduleSlots = (moduleSlots < 1)? 1 : moduleSlots;
		_modules = new Vector.<IGrayBoxModule>(_moduleSlots, true);
		_bypassFlag = new Vector.<Boolean>(_moduleSlots, true);
		for (var i:int = 0; i < _moduleSlots; i++)
		{
			_modules[i] = null;
			_bypassFlag[i] = false;
		}
	}
}
}