/*! ********************************************
 * MidiMessage.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox 
{
import flash.utils.ByteArray;

public class MidiMessage 
{
/*
	[ Channel Message ]
		NoteOff
			type    = 0x80
			channel = 0 - 15
			value0  = 0 - 127   NoteNumber
			value1  = 0
			data    = null
			
		NoteOn
			type    = 0x90
			channel = 0 - 15
			value0  = 0 - 127   NoteNumber
			value1  = 0 - 127   Velocity
			data    = null
			
		Key AfterTouch
			type    = 0xA0
			channel = 0 - 15
			value0  = 0 - 127   NoteNumber
			value1  = 0 - 127   Value
			data    = null
			
		Control Change
			type    = 0xB0
			channel = 0 - 15
			value0  = 0 - 127   CC Number
			value1  = 0 - 127   Value
			data    = null
			
		Program Change
			type    = 0xC0
			channel = 0 - 15
			value0  = 0 - 127   ProgramNumber
			value1  = 0
			data    = null
			
		Channel AfterTouch
			type    = 0xD0
			channel = 0 - 15
			value0  = 0 - 127   Value
			value1  = 0
			data    = null
			
		PitchBend
			type    = 0xE0
			channel = 0 - 15
			value0  = 0 - 127   Value  LSB
			value1  = 0 - 127   Value  MSB
			data    = null
			
	[ System Common Message ]
		System Exclusive (*)
			type    = 0xF0
			channel = 0
			value0  = 0
			value1  = 0
			data    = data... F7[EOX]  (SysEx Message)
			
		System Exclusive (*)
			type    = 0xF7
			channel = 0
			value0  = 0
			value1  = 0
			data    = data... F7[EOX]  (SysEx Message)
			
			(*) Avoid using Multi-Packet SysExMessage, if possible..
				SysEx Event 1 (Packet 1) : F0  data....
				SysEx Event 2 (Packet 2) : F7  data....
				SysEx Event 3 (Packet 3) : F7  data....F7[EOX]
		
		MIDI Time Code 
			type    = 0xF1
			channel = 0
			value0  = 0 - 127
			value1  = 0
			data    = null
			
		Song Position
			type    = 0xF2
			channel = 0
			value0  = 0 - 127   Value  LSB
			value1  = 0 - 127   Value  MSB
			data    = null
			
		Song Select
			type    = 0xF3
			channel = 0
			value0  = 0 - 127   SongNumber
			value1  = 0
			data    = null
			
		Tune Request
			type    = 0xF6
			channel = 0
			value0  = 0
			value1  = 0
			data    = null
			
	[ System RealTime Message ]
			type    = 0xF8 ~ 0xFF
			channel = 0
			value0  = 0
			value1  = 0
			data    = null
			
	[ Meta Message (SMF Track Event) ]
		type    = MetaType [0xFF00 - 0xFF7F]
		channel = 0
		value0  = 0 or value(*)
		value1  = 0
		data    = data... (Meta Event Data)
		
		(*)
			FF00  sequence number [0 - 65535]
			FF20  channel prefix  [0 - 15]
			FF2F  end of track    [0]
			FF51  set tempo       [0 - 16777215]
*/
// Mesassage Data [uint8]
	public var type    :int = 0x80; // * May contain NonStandard MidiMessages.
	public var channel :int = 0;
	public var value0  :int = 0;
	public var value1  :int = 0;
	
	public var data    :ByteArray = null;
	//public var dataSize :int = 0;
	
	public function clear():void
	{
		type = 0x80;
		channel = 0;
		value0 = 0;
		value1 = 0;
		data = null;
		//dataSize = 0;
	}
	
// Constructor
	public function MidiMessage():void{}
}
}