/*! ********************************************
 * VoiceUnitPCM.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth 
{
internal final class VoiceUnitPCM extends VoiceUnitBase
{
// Tables
	// PG
	private const _PG_PHASE_FP      :int = SynthesizerPCM.PG_PHASE_FP;
	private const _PG_PHASE_FP_MASK :int = (1 << _PG_PHASE_FP) - 1;
	// LFO
	private const _LFO_TIMER_INIT_VALUE :int = SynthData.LFO_TIMER_INIT_VALUE;
	private const _LFO_WAVE_TABLE_MASK  :int = SynthData.LFO_WAVE_TABLE_SIZE - 1;
	private var   _LFO_WAVE_TABLE       :Vector.<int> = null;
// Synthesizer Data
	internal var VOICE_PARTS     :Vector.<VoicePart> = null;
	internal var SAMPLE_DATAS    :Vector.<SampleData> = null;
	internal var TONE_PRESETS    :Vector.<TonePresetPCM> = null;
	internal var SAMPLER_PRESETS :Vector.<SamplerPreset> = null;
	internal var pgSamplePow2tIdx :int = 0;  // PG PhaseStep (Pow2Table Index Offset) - SamplingRate
	internal var pgKeyPow2tIdx    :int = 0;  // PG PhaseStep (Pow2Table Index Offset) - ReferenceKey (NoteNumber 0)
	internal var egPow2tIdx       :int = 0;  // EG PhaseStep (Pow2Table Index Offset)
	internal var egPhaseStepSndOff:int = 0;
	internal var lfoTimerCoef     :int = 0;  // Coefficient Value - LFO Phase Update Timer
	// Pitch
	private var _ptBend     :int = 0;   // pitchBend
	private var _ptBendSens :int = 0;   // pitchBend Sensitivity
	// LFO
	private var _lfoPhase     :int = 0;  // LFO Phase
	private var _lfoTimer     :int = 0;  // LFO Phase Update Timer
	private var _lfoTimerStep :int = 0;  // Update Timer Step
	private var _lfoMod       :int = 0;  // Modulation Depth
	private var _lfoModSens   :int = 0;  // Modulation Sensitivity
	private var _lfoOutPM     :int = 0;  // output - phase modulation
	private var _lfoOutAM     :int = 0;  // output - amplitude modulation
	// Wave
	private var _waveTable :Vector.<int> = null;
	// EG
	/*
	- Envelop Generator State -
	[ 5674 3210 ] 8Bits Flag Status (0 = deActive)
	[ ---- RSDA ] "-" = Reserved Bits
		A = Attack  stage
		D = Decay   stage
		S = Sustain stage
		R = Release stage
	*/
	private var _egState      :int = 0;
	private var _egLevelTable :Vector.<int> = null;
	private var _egCount      :int = 0;
	private var _egCountStep  :int = 0;
	//private var _egCountStepAR:int = 0;
	private var _egCountStepDR:int = 0;
	private var _egCountStepRR:int = 0;
	private var _egCntEnvSus  :int = 0;
	private var _egCntEnvBtm  :int = 0;
	private var _egGate       :int = 0;
	private var _egSL         :int = 0;
	private var _egTL         :int = 0;
	private var _egTLv        :int = 0;
	// PG
	private var _pgKey        :int = 0;
	private var _pgBend       :int = 0;
	private var _pgPhase      :int = 0;
	private var _pgPhaseAFP   :int = 0;  // PG Phase (After the Fixed Point)
	private var _pgPhaseStep  :int = 0;
// Sampler
	private var _plbkE      :int = 0;  // PlayBack End Frame   [plbkE <= channelPerSamples]
	private var _loopE      :int = 0;  // Loop End Frame       [loopE <= plbkE]
	private var _loopF      :int = 0;  // Loop Frames          [val = loopE - lpStart]
	private var _loopFk     :int = 0;  // Loop Frames          * after -> keyOff
	
	override internal function keyOff():void
	{
		// KeyOn --->
		if (state & 0x0010)
		{
			// ATTACK ->
			if (_egState == 1)
			{
				_egLevelTable = _EG_LEVEL_TABLE_DW;
				if (_egCount >= _EG_MAX_CNT)_egCount = 0;
				else _egCount = _EG_LEVEL_TABLE_UP[_egCount >> _EG_CNT_FP] << _EG_CNT_FP - 1;  // -1 = 8192 to 4096
				if (_egCount > _egCntEnvBtm){ killProcess(); return; }
			}
			// DECAY,SUSTAIN ->
			_egGate = _egCntEnvBtm;
			_egCountStep = _egCountStepRR;
			_egState = 8;  // Release
			state = 0x0021;
			// sampler ---
			_loopE = _plbkE;
			_loopF = _loopFk;
		}
	}
	override internal function soundOff():void
	{
		// KeyOn || KeyOff --->
		if (state & 0x0030)
		{
			// ATTACK ->
			if (_egState == 1)
			{
				_egLevelTable = _EG_LEVEL_TABLE_DW;
				if (_egCount >= _EG_MAX_CNT)_egCount = 0;
				else _egCount = _EG_LEVEL_TABLE_UP[_egCount >> _EG_CNT_FP] << _EG_CNT_FP - 1;
				if (_egCount > _egCntEnvBtm){ killProcess(); return; }
			}
			// DECAY,SUSTAIN ->
			_egGate = _egCntEnvBtm;
			_egCountStep = egPhaseStepSndOff;
			_egState = 0;  // SoundOff, EG OFF
			partNumber = -1; // SoundOff
			state = 0x0041;
			// sampler ---
			_loopE = _plbkE;
			_loopF = _loopFk;
		}
	}
	private function _shiftEGState():int
	{
		switch (_egState)
		{
			// ATTACK ->
			case 1:
				_egLevelTable = _EG_LEVEL_TABLE_DW;
				_egCount = 0;
				// -> DECAY
				if (_egSL > 0)
				{
					_egGate = _egCntEnvSus;
					_egCountStep = _egCountStepDR;
					return _egState = 2;
				}
				// -> SUSTAIN
				_egGate = 0x7fffffff;
				_egCountStep = 0;
				return _egState = 4;
			// DECAY ->
			case 2:
				// -> SUSTAIN
				if ((_egTL + _egSL) < _EG_SUSTAIN_BTM)
				{
					_egLevelTable = _EG_LEVEL_TABLE_DW;
					_egGate = 0x7fffffff;
					_egCountStep = 0;
					_egCount = _egCntEnvSus;
					return _egState = 4;
				}
			// DECAY,RELEASE,SoundOFF -> EG OFF
			default: return _egState = 0;
		}
	}
	override internal function keyOn(ch:int, note:int, velo:int):void
	{
	// KeyOn Status ---
		state = 0x0011;
		partNumber = ch;
		noteNumber = note;
	// Part, Preset ---
		var _part:VoicePart = VOICE_PARTS[partNumber], 
			_preset:TonePresetPCM = TONE_PRESETS[_part.programChange],
			_sample:SampleData = SAMPLE_DATAS[_preset.smplNum];
		if (_sample.sampleData == null){ killProcess(); return; }
		_waveTable = _sample.sampleData;
		_plbkE = _sample.plbkE;
		_loopE = _sample.loopE;
		_loopF = _sample.loopF;
		_loopFk = _sample.loopFk;
	// LFO ---
		_lfoModSens = _preset.lms;
		_lfoMod = _part.ccValues[1] * _lfoModSens;
		_lfoTimerStep = _preset.lfoFreq * lfoTimerCoef;
		_lfoTimer = _LFO_TIMER_INIT_VALUE;
		_lfoPhase = _preset.strc + 1 & 1;
		_lfoOutPM = 0;
		_lfoOutAM = 0;
	// volume ---
		_vPtVol = _mainVolTable[_part.ccValues[7]];
		_vPtExp = _exprVolTable[_part.ccValues[11]];
		_vVel = _veloVolTable[velo];
		_vVol = _vVel * _masterVolume;
		_vOut = _vPtVol * _vPtExp * _vVol;
		_prPan = 64;
		_vPanL = _panTable[      _prPan + _part.ccValues[10]];
		_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	// PG ---
		_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
		_ptBend = _part.pitchBend - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
		_pgKey = _POW2T_KEY_TABLE[noteNumber] + (_preset.tune - 8192) + _sample.pgKey + pgKeyPow2tIdx;
		var v:int = _pgKey + _pgBend;// & 0x7fffffff;
		_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase = _sample.plbkS;
		_pgPhaseAFP = 0;
	// EG ---
		_egTL = _preset.tl;
		_egTLv = _EG_TOTAL_LEVEL_TABLE[_egTL];
		_egCntEnvBtm = _EG_CNT_GATE_TABLE[128 - _egTL];
		_egSL = _preset.sl;
		_egCntEnvSus = _EG_CNT_GATE_TABLE[_egSL];
		if (_egCntEnvSus > _egCntEnvBtm)_egCntEnvSus = _egCntEnvBtm;
		var _keyFlw:int = _EG_KEY_TABLE[127 + noteNumber - _preset.kfStart] << _preset.kf;
		_keyFlw += egPow2tIdx;
		v = (_preset.dr << _EG_PRAM_OCTV_LS) + _keyFlw;
		_egCountStepDR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset.rr << _EG_PRAM_OCTV_LS) + _keyFlw;
		_egCountStepRR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	// EG State ---
		// -> ATTACK
		if (_preset.ar < 127) {
			_egCount = _EG_CNT_ENV_ATK;
			_egLevelTable = _EG_LEVEL_TABLE_UP;
			_egGate = _EG_MAX_CNT;
			v = (_preset.ar << _EG_PRAM_OCTV_LS) + _keyFlw;
			_egCountStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState = 1;
		} else {
			_egLevelTable = _EG_LEVEL_TABLE_DW;
			_egCount = 0;
		// -> DECAY
			if (_egSL > 0) {
				_egGate = _egCntEnvSus;
				_egCountStep = _egCountStepDR;
				_egState = 2;
		// -> SUSTAIN
			} else {
				_egGate = 0x7fffffff;  // dummy
				_egCountStep = 0;  // dummy
				_egState = 4; }
		}
	// process ---
		v = (_preset.strc & 3) << 1;
		if (_sample.channelCount == 2)v|=1;
		process = _PROC_PTRS[v];
	}
	internal function keyOnSampler(ch:int, note:int, velo:int):void
	{
	// KeyOn Status ---
		state = 0x0011;
		partNumber = ch;
		noteNumber = note;
	// tone ---
		var _part:VoicePart = VOICE_PARTS[partNumber],
			_preset:SamplerPreset = SAMPLER_PRESETS[noteNumber],
			_sample:SampleData = SAMPLE_DATAS[_preset.smplNum];
		if (_sample.sampleData == null){ killProcess(); return; }
		_waveTable = _sample.sampleData;
		_plbkE = _sample.plbkE;
		_loopE = _sample.loopE;
		_loopF = _sample.loopF;
		_loopFk = _sample.loopFk;
	// LFO ---
		_lfoOutPM = 0;
		_lfoOutAM = 0;
	// volume ---
		_vPtVol = _mainVolTable[_part.ccValues[7]];
		_vPtExp = _exprVolTable[_part.ccValues[11]];
		_vVel = _veloVolTable[velo];
		_vVol = _vVel * _masterVolume;
		_vOut = _vPtVol * _vPtExp * _vVol;
		_prPan = _preset.pan;
		_vPanL = _panTable[      _prPan + _part.ccValues[10]];
		_vPanR = _panTable[256 - _prPan - _part.ccValues[10]];
		_vOutL = _vOut * _vPanL;
		_vOutR = _vOut * _vPanR;
	// PG ---
		_ptBendSens = _POW2T_KEY_TABLE[_part.bendSens];
		_ptBend = _part.pitchBend - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
		_pgKey = (_preset.octv - 8 << 13) + (_preset.tune - 8192) + _sample.pgSample + pgSamplePow2tIdx;
		var v:int = _pgKey + _pgBend;
		_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		_pgPhase = _sample.plbkS;
		_pgPhaseAFP = 0;
	// EG ---
		_egTL = _preset.tl;
		_egTLv = _EG_TOTAL_LEVEL_TABLE[_egTL];
		_egCntEnvBtm = _EG_CNT_GATE_TABLE[128 - _egTL];
		_egSL = _preset.sl;
		_egCntEnvSus = _EG_CNT_GATE_TABLE[_egSL];
		if (_egCntEnvSus > _egCntEnvBtm)_egCntEnvSus = _egCntEnvBtm;
		v = (_preset.dr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
		_egCountStepDR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
		v = (_preset.rr << _EG_PRAM_OCTV_LS) + egPow2tIdx;
		_egCountStepRR = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	// EG State ---
		// -> ATTACK
		if (_preset.ar < 127) {
			_egCount = _EG_CNT_ENV_ATK;
			_egLevelTable = _EG_LEVEL_TABLE_UP;
			_egGate = _EG_MAX_CNT;
			v = (_preset.ar << _EG_PRAM_OCTV_LS) + egPow2tIdx;
			_egCountStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _EG_POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
			_egState = 1;
		} else {
			_egLevelTable = _EG_LEVEL_TABLE_DW;
			_egCount = 0;
		// -> DECAY
			if (_egSL > 0) {
				_egGate = _egCntEnvSus;
				_egCountStep = _egCountStepDR;
				_egState = 2;
		// -> SUSTAIN
			} else {
				_egGate = 0x7fffffff;  // dummy
				_egCountStep = 0;  // dummy
				_egState = 4; }
		}
	// process ---
		process = (_sample.channelCount == 2)? _procSmpl2ch: _procSmpl1ch;
	}
	override internal function updPartBendSens(val:int):void
	{
		_ptBendSens = _POW2T_KEY_TABLE[val];
		_pgBend = _ptBend * _ptBendSens >> 13;
		var v:int = _pgKey + _pgBend + _lfoOutPM;
		_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	}
	override internal function updPartPitchBend(val:int):void
	{
		_ptBend = val - 8192;
		_pgBend = _ptBend * _ptBendSens >> 13;
		var v:int = _pgKey + _pgBend + _lfoOutPM;
		_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
	}
	override internal function updPartModulation(val:int):void
	{
		_lfoMod = val * _lfoModSens;
	}
	internal function updateLfoFrequency(val:int):void
	{
		_lfoTimerStep = val * lfoTimerCoef;
	}
// Constructor
	public function VoiceUnitPCM():void
	{
		super();
		var _synthData:SynthData = SynthData.getInstance();
		// Log2 Volume Table
		_LOG2V_TABLE = _synthData.getLOG2V_TABLE();
		// Pow2 Table
		_POW2_TABLE      = _synthData.getPOW2_TABLE();
		_POW2T_RS_TABLE  = _synthData.getPOW2T_RS_TABLE();
		_POW2T_KEY_TABLE = _synthData.getPOW2T_KEY_TABLE();
		// EG
		_EG_TOTAL_LEVEL_TABLE = _synthData.getEG_TOTAL_LEVEL_TABLE();
		_EG_LEVEL_TABLE_UP    = _synthData.getEG_LEVEL_TABLE_UP();
		_EG_LEVEL_TABLE_DW    = _synthData.getEG_LEVEL_TABLE_DW();
		_EG_KEY_TABLE         = _synthData.getEG_KEY_TABLE();
		_EG_POW2T_RS_TABLE    = _synthData.getEG_POW2T_RS_TABLE();
		_EG_CNT_GATE_TABLE    = _synthData.getEG_CNT_GATE_TABLE();
		// etc
		_LFO_WAVE_TABLE = _synthData.getLFO_WAVE_TABLES()[0];
		
		_PROC_PTRS[0] = _proc0;
		_PROC_PTRS[1] = _proc1;
		_PROC_PTRS[2] = _proc2;
		_PROC_PTRS[3] = _proc3;
		_PROC_PTRS[4] = _proc4;
		_PROC_PTRS[5] = _proc5;
	}
// process ---
	// pointer array (process Function)
	private const _PROC_PTRS :Vector.<Function> = new Vector.<Function>(6, true);
	// Inst 1ch
	private function _proc0(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, n:Number, v:int,
			_et:Vector.<int> = _egLevelTable;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, 
			_wt:Vector.<int> = _waveTable;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState())_et = _egLevelTable;
				else { killProcess(); return; }
			}
			v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
			v = _wt[_pgPhase] + v;
			n = _log2vT[v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Inst 2ch
	private function _proc1(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, v:int, p:int, 
			_et:Vector.<int> = _egLevelTable;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, 
			_wt:Vector.<int> = _waveTable;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState())_et = _egLevelTable;
				else { killProcess(); return; }
			}
			v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
		// out --->
			p = _pgPhase << 1;
			outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
			outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
		// counter --->
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Inst 1ch + LFO PM
	private function _proc2(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, n:Number, v:int,
			_et:Vector.<int> = _egLevelTable;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, 
			_wt:Vector.<int> = _waveTable;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				_lfoOutPM = (_LFO_WAVE_TABLE[_lfoPhase] * _lfoMod) >> 11;  // 7(mod) + 4(sens)
				v = _pgKey + _pgBend + _lfoOutPM;
				_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState())_et = _egLevelTable;
				else { killProcess(); return; }
			}
			v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
			v = _wt[_pgPhase] + v;
			n = _log2vT[v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Inst 2ch + LFO PM
	private function _proc3(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, v:int, p:int,
			_et:Vector.<int> = _egLevelTable;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, 
			_wt:Vector.<int> = _waveTable;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				_lfoOutPM = (_LFO_WAVE_TABLE[_lfoPhase] * _lfoMod) >> 11;  // 7(mod) + 4(sens)
				v = _pgKey + _pgBend + _lfoOutPM;
				_pgPhaseStep = _POW2_TABLE[v & _POW2T_OCTV_MASK] >> _POW2T_RS_TABLE[v >> _POW2T_OCTV_BITS];
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState())_et = _egLevelTable;
				else { killProcess(); return; }
			}
			v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
		// out --->
			p = _pgPhase << 1;
			outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
			outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Inst 1ch + LFO AM
	private function _proc4(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, n:Number, v:int,
			_et:Vector.<int> = _egLevelTable;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, 
			_wt:Vector.<int> = _waveTable;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				v = (_LFO_WAVE_TABLE[_lfoPhase] * _lfoMod) >> 11;
				_lfoOutAM = v & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState())_et = _egLevelTable;
				else { killProcess(); return; }
			}
			v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
			v = _wt[_pgPhase] + v + _lfoOutAM;
			n = _log2vT[v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Inst 2ch + LFO AM
	private function _proc5(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, v:int, p:int,
			_et:Vector.<int> = _egLevelTable;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, 
			_wt:Vector.<int> = _waveTable;
		for ( ; i < end; i++)
		{
		// lfo Update --->
			if (_lfoTimer <= 0)
			{
				v = (_LFO_WAVE_TABLE[_lfoPhase | 1] * _lfoMod) >> 11;
				_lfoOutAM = v & 0x7ffffffe;
				_lfoPhase = _lfoPhase + 2 & _LFO_WAVE_TABLE_MASK;
				_lfoTimer += _LFO_TIMER_INIT_VALUE;
			}
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState())_et = _egLevelTable;
				else { killProcess(); return; }
			}
			v = _et[_egCount >> _EG_CNT_FP] + _egTLv + _lfoOutAM;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
		// out --->
			p = _pgPhase << 1;
			outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
			outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
		// counter --->
			_lfoTimer -= _lfoTimerStep;
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Sampler 1ch
	private function _procSmpl1ch(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, n:Number, v:int,
			_et:Vector.<int> = _egLevelTable;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, 
			_wt:Vector.<int> = _waveTable;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState())_et = _egLevelTable;
				else { killProcess(); return; }
			}
			v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
			v = _wt[_pgPhase] + v;
			n = _log2vT[v];
		// out --->
			outL[i] += n * _vOutL;
			outR[i] += n * _vOutR;
		// counter --->
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
	// Sampler 2ch
	private function _procSmpl2ch(outL:Vector.<Number>, outR:Vector.<Number>, end:int):void
	{
		var i:int = startFrame, v:int, p:int, 
			_et:Vector.<int> = _egLevelTable;
		const _log2vT:Vector.<Number> = _LOG2V_TABLE, 
			_wt:Vector.<int> = _waveTable;
		for ( ; i < end; i++)
		{
		// eg update --->
			if (_egCount >= _egGate)
			{
				if (_shiftEGState())_et = _egLevelTable;
				else { killProcess(); return; }
			}
			v = _et[_egCount >> _EG_CNT_FP] + _egTLv;	
		// pg update --->
			if (_pgPhase >= _loopE)
			{
				if (_loopF > 0) do{ _pgPhase -= _loopF; } while (_pgPhase >= _loopE);
				else { killProcess(); return; }
			}
		// out --->
			p = _pgPhase << 1;
			outL[i] += _log2vT[_wt[p    ] + v] * _vOutL;
			outR[i] += _log2vT[_wt[p + 1] + v] * _vOutR;
		// counter --->
			_egCount += _egCountStep;
			_pgPhaseAFP += _pgPhaseStep;
			_pgPhase += _pgPhaseAFP >> _PG_PHASE_FP;
			_pgPhaseAFP &= _PG_PHASE_FP_MASK;
		}
		startFrame = i;
	}
}
}