/*! ********************************************
 * Main.as
 * Copyright 2022 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> license.txt
 ******************************************** */
package
{
	import flash.display.Sprite;
	import flash.events.SampleDataEvent;
	import flash.media.Sound;
	import flash.utils.ByteArray;
	import sora.graybox.GrayBox;
	
	public class Main extends Sprite 
	{
		private var _sound:Sound;
		private var _audioBuffer:Vector.<Vector.<Number>>;
		private var _graybox:GrayBox;
		private var _noteGen:NoteGenerator;
		private var _sineGen:SineGenerator;
		
		// ストリーミング処理 -->
		private function _sampleDataEvent(e:SampleDataEvent):void
		{
			var i:int = 0;
		// オーディオバッファのクリア -->
			var _buffL:Vector.<Number> = _audioBuffer[0], // チャンネル L
			    _buffR:Vector.<Number> = _audioBuffer[1]; // チャンネル R
			for (i = 0; i < 2048; i++)
			{
				_buffL[i] = 0.0;
				_buffR[i] = 0.0;
			}
			
		// GrayBoxによるオーディオ、イベント処理 -->
			// GrayBox.process()メソッドを使用します。
			// 第1引数 : 書き込みを行うオーディオバッファ
			// 第2引数 : 書き込み開始フレーム
			// 第3引数 : 書き込みフレーム数
			// ・今回はオーディオバッファ全体を使うので、開始フレームが0、書き込みフレーム数が2048
			// ・スロットナンバー0から順にモジュール側のIGrayBoxModule.process()メソッドが呼ばれるので、
			//   モジュール側で音声、イベント処理を行います。
			_graybox.process(_audioBuffer, 0, 2048);
			
		// オーディオバッファからオーディオストリームへ書き込み -->
			var _stream:ByteArray = e.data;
			for (i = 0; i < 2048; i++)
			{
				_stream.writeFloat(_buffL[i]);
				_stream.writeFloat(_buffR[i]);
			}
		}
		
		public function Main():void
		{
		// ストリーミング再生を行うオブジェクト生成 -->
			_sound = new Sound();
		// オーディオバッファ生成 -->
			// ブロックサイズ2048フレーム
			_audioBuffer = new Vector.<Vector.<Number>>(2, true);
			_audioBuffer[0] = new Vector.<Number>(2048, true); // チャンネル L
			_audioBuffer[1] = new Vector.<Number>(2048, true); // チャンネル R
		// GrayBoxの構築 -->
			// 第1引数 : ブロックサイズ         ( 1~ ) Flash環境は2048~8192フレーム
			// 第2引数 : サンプリングレート[Hz] ( 1~ ) Flash環境は44100Hz固定
			// 第3引数 : オーディオチャンネル数 ( 2~ )
			// 第4引数 : モジュールスロット数   ( 1~ ) 今回は2つのモジュールを使うので2
			// ・初期値は必須パラメーターですが、構築時に確定できない場合は暫定値で構いません。
			//   ただし、ストリーミング前には設定値を確定させてください。
			_graybox = new GrayBox(2048, 44100.0, 2, 2);
			
		// 各モジュールの構築 -->
			// 44100フレーム(1秒)間隔でドレミのMidiメッセージを出力し続けるモジュール
			_noteGen = new NoteGenerator();
			// Midiメッセージを元にサイン波を出力するモジュール
			_sineGen = new SineGenerator();
			
		// モジュールの接続 -->
			// GrayBox.attachModule()メソッドを使用し、モジュールを接続します。
			// 第1引数 : 接続するスロットのインデックスナンバー
			// 第2引数 : 接続するモジュールオブジェクト
			// 接続されたモジュールはインデックスナンバー0から順に処理されます。
			// ・正常に接続されると、モジュール側のIGrayBoxModule.connectModule()メソッドが呼ばれるので、
			//   こちらから初期化などを行うことができます。
			_graybox.attachModule(0, _noteGen);
			_graybox.attachModule(1, _sineGen);
			
			
		// ストリーミングの開始 -->
			// ストリーミング前処理
			// ・モジュール側IGrayBoxModule.startProcess()メソッドが呼ばれます。
			//   必要に応じてモジュール側で前処理を行えます。
			_graybox.startProcess();
			// ストリーミング開始
			_sound.addEventListener(SampleDataEvent.SAMPLE_DATA, _sampleDataEvent);
			_sound.play();
			
		}
	}
}