/*! ********************************************
 * NoteGenerator.as
 * Copyright 2022 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> license.txt
 ******************************************** */
package 
{
	import sora.graybox.GrayBox;
	import sora.graybox.IGrayBoxModule;
	import sora.graybox.MidiEvent;
	import sora.graybox.ProcessData;
	
public class NoteGenerator implements IGrayBoxModule 
{
	private var _crntKey:int = 0;
	private var _prevKey:int = 0;
	private var _phase:int = 0;
	private var _sampleRate:int = 44100;
	
	public function NoteGenerator():void {}
// sora.graybox.IGrayBoxModule ---
	public function updateSampleRate(sampleRate:Number):void
	{ 
		_sampleRate = int(sampleRate);
		if (_phase >= _sampleRate)_phase = 0;
	}
	public function updateAudioChannels(ch:int):void{}
	public function updateProcessBlockSize(size:int):void{}
	public function resetModule():void{ _phase = 0; _crntKey = 0; _prevKey = 12; }
	public function initiModule():void{ _phase = 0; _crntKey = 0; _prevKey = 12; }
	public function connectModule(host:GrayBox):void
	{
		updateSampleRate(host.getSampleRate());
	}
	public function disconnectModule():void{ }
	public function updateBypassStatus(bypass:Boolean):void{}
	public function startProcess():void{}
	public function stopProcess():void{}
	public function suspendProcess():void{}
	public function resumeProcess():void{}
	
	public function process(procData:ProcessData):void 
	{
		var _noteNumber:int, _midiEvent:MidiEvent;
		// 開始フレーム
		var _frame:int = procData.offsetFrames;
		// 終了フレーム
		var _frameEnd:int = _frame + procData.sampleFrames;
		
	// Midiイベント生成 -->
		while (_frame < _frameEnd)
		{
			// 1秒ごとにMidiイベント生成
			if (_phase++ >= _sampleRate)
			{
			// ノートオフメッセージ (前回の分)
				_noteNumber = _prevKey + 72;
				_midiEvent = new MidiEvent();
				_midiEvent.type = 0x80;           // ノートオフ
				_midiEvent.channel = 0;           // Midiチャンネル
				_midiEvent.value0 = _noteNumber;  // ノートナンバー
				_midiEvent.offsetFrames = _frame; // このイベントの発生フレーム
				procData.addToMidiEventList(_midiEvent); // Midiイベントリストへ登録
			// ノートオンメッセージ (今回の分)
				_noteNumber = _crntKey + 72;
				_midiEvent = new MidiEvent();
				_midiEvent.type = 0x90; // ノートオン
				_midiEvent.channel = 0; // Midiチャンネル
				_midiEvent.value0 = _noteNumber; // ノートナンバー
				_midiEvent.value1 = 127; // ノートオンベロシティー
				_midiEvent.offsetFrames = _frame; // このイベントの発生フレーム
				procData.addToMidiEventList(_midiEvent); // Midiイベントリストへ登録
				
				if (++_crntKey > 12)_crntKey = 0;
				if (++_prevKey > 12)_prevKey = 0;
				_phase = 0;
			}
			_frame++;  // フレーム更新
		}
	}
}
}