/*! ********************************************
 * SineGenerator.as
 * Copyright 2022 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> license.txt
 ******************************************** */
package 
{
	import sora.graybox.GrayBox;
	import sora.graybox.IGrayBoxModule;
	import sora.graybox.MidiEvent;
	import sora.graybox.ProcessData;
	
public class SineGenerator implements IGrayBoxModule 
{
	public var volume:Number = 0.3;
	private var _keyOn:Boolean = false;
	private var _frequency:Number = 440.0;
	private var _sampleRate:Number = 44100.0;
	private var _frames:Number = 0.0;
	private var _phase:int = 0;
	
	public function SineGenerator():void {}
// sora.graybox.IGrayBoxModule ---
	public function updateSampleRate(sampleRate:Number):void
	{
		_sampleRate = sampleRate;
		_frames = _sampleRate / _frequency;
	}
	public function updateAudioChannels(ch:int):void{}
	public function updateProcessBlockSize(size:int):void{}
	public function resetModule():void{ volume = 0.3; _keyOn = false; }
	public function initiModule():void{ volume = 0.3; _keyOn = false; }
	public function connectModule(host:GrayBox):void
	{ 
		_keyOn = false;
		updateSampleRate(host.getSampleRate());
	}
	public function disconnectModule():void{}
	public function updateBypassStatus(bypass:Boolean):void{ if (bypass)_keyOn = false; }
	public function startProcess():void{ _keyOn = false; }
	public function stopProcess():void{ _keyOn = false; }
	public function suspendProcess():void{}
	public function resumeProcess():void{}
	
	public function process(procData:ProcessData):void 
	{
		// 開始フレーム
		var _frame:int = procData.offsetFrames;
		// 終了フレーム
		var _frameEnd:int = _frame + procData.sampleFrames;
		
	// Midiイベント処理 -->
		// ProcessData.midiEvents (Midiイベントリスト(片方向リンクリスト))
		// ・ProcessData.addToMidiEventList()メソッドで追加したMidiイベントはこのリストに追加される
		// ・ProcessData.midiEvents自体はリスト管理用オブジェクトなので、
		//   イベント処理はProcessData.midiEvents.next以降を使用してください。
		var _event:MidiEvent = procData.midiEvents.next;
		// リストからMidiイベントを順に読み込み、処理を行う
		while (_event != null)
		{
			// このMidiイベントまで音声処理
			if (_keyOn)_sineGen(procData.audioOut, _frame, _event.offsetFrames);
			
			// このMidiイベントのタイプ
			switch (_event.type)
			{
				case 0x80: // ノートオフ
					_frame = _event.offsetFrames;  // 開始フレーム更新
					_keyOn = false;  // 音声処理フラグ オフ
					break;
					
				case 0x90: // ノートオン
					_frame = _event.offsetFrames; // 開始フレーム更新
					_keyOn = true;  // 音声処理フラグ オン
					// 周波数などの更新
					_frequency = 440.0 * Math.pow(2.0, (_event.value0 - 69.0) / 12.0);
					_frames = _sampleRate / _frequency;
					_phase = 0;     // 位相のリセット
					break;
			}
			// 次のMidiイベント
			_event = _event.next;
		}
		
		// 終了フレームまで音声処理
		if (_keyOn)_sineGen(procData.audioOut, _frame, _frameEnd);
	}
	
	private function _sineGen(out:Vector.<Vector.<Number>>, startFrame:int, frameEnd:int):void
	{
		var _outL:Vector.<Number> = out[0],
		    _outR:Vector.<Number> = out[1];
			
		for (var i:int = startFrame; i < frameEnd; i++)
		{
			var n:Number = Math.sin((_phase / _frames) * 3.141592653589793) * volume;
			_phase++;// 位相の更新
			_outL[i] += n;
			_outR[i] += n;
		}
	}
}
}